@@ -1,3 +1,5 @@
+#![allow(dead_code)]
+
 use ron::de::from_reader;
 use serde::Deserialize;
 use std::{collections::HashMap, fs::File};
@@ -9,6 +11,7 @@ struct Config {
     map: HashMap<u8, char>,
     nested: Nested,
     tuple: (u32, u32),
+    vec: Vec<Nested>,
 }
 
 #[derive(Debug, Deserialize)]
